#include <iostream>
#include <vector>
#include <map>
#include <algorithm>
#include "graph.hpp"
#include "../solver/Solver.hpp"

/**
#define TIMEWAIT 3            // le temps d'attente minimum dans une gare desservie
#define SLOW 2                // nombre de train lent
#define FAST 1                // nombre de train rapide
#define TRAIN (SLOW+FAST)     // nombre de train
#define TIMESLOT 26           // nombre de minutes dans la plage horaire
#define STATION 3             // nombre de gare (voir la map)
#define TIMEDURATION 3        // durée maximal des voyages
#define TIMEWINDOW 10         // fréquence des trains direct
#define FILE "../maps/slow-fast.txt"
*/



#define TIMEWAIT 3            // le temps d'attente minimum dans une gare desservie
#define SLOW 1                // nombre de train lent
#define FAST 0                // nombre de train rapide
#define TRAIN (SLOW+FAST)     // nombre de train
#define TIMESLOT 26           // nombre de minutes dans la plage horaire
#define STATION 4             // nombre de gare (voir la map)
#define TIMEDURATION 12       // durée maximal des voyages
#define TIMEWINDOW 13         // fréquence des trains direct
#define FILE "../maps/cycle.txt"


int G[TRAIN][TIMESLOT][STATION];
int V[TRAIN][TIMESLOT][STATION][STATION];
int J[TRAIN][TIMESLOT][TIMESLOT][STATION][STATION];
Graph *map;
Solver solver;
vec<Lit> literals;
std::vector<int> slowTrainsList;
std::vector<int> fastTrainsList;
std::vector<int> allTrains;
std::vector<std::vector<int>> allPairStationList;

/**
 * Check if there is a connection between two trains stations
 * @param from train station start
 * @param to  train station end
 * @return
 */
bool isConnected(int from, int to) {
    return map->get_duration(from, to) > 0;
}

/**
 * Check if the station is a small station
 * @param node
 * @return
 */
bool isSmallStation(int node) {
    return map->get_type(node) == Small;
}

/**
 * Check if the train is a fast train
 * @param train
 * @return
 */
bool isFastTrain(int train) {
    return std::find(fastTrainsList.begin(), fastTrainsList.end(), train) != fastTrainsList.end();
}

/**
 * initialize variable G which represents the fact that a train is in a station at a certain time
 */
void init_VarTrainInStationAtTime() {
    for (int train : allTrains)
        for (int time = 0; time < TIMESLOT; time++)
            for (int station = 0; station < STATION; station++)
                G[train][time][station] = solver.newVar();
}

/**
 * initialize variable V which represents the fact that a train is on a track between two station at a certain time
 */
void init_VarTrainOnTrackAtTime() {
    for (int train: allTrains)
        for (int time = 0; time < TIMESLOT; time++)
            for (int stationFrom = 0; stationFrom < STATION; stationFrom++)
                for (int stationTo = 0; stationTo < STATION; stationTo++)
                    V[train][time][stationFrom][stationTo] = solver.newVar();
}

/**
 * initialize variable J which represents the fact that a train is on a journey between two station at a certain time
 */
void init_VarJourney() {
    for (int train: allTrains)
        for (int time = 0; time < TIMESLOT; time++)
            for (int time2 = 0; time2 < TIMESLOT; time2++)
                for (int stationFrom = 0; stationFrom < STATION; stationFrom++)
                    for (int stationTo = 0; stationTo < STATION; stationTo++)
                        J[train][time][time2][stationFrom][stationTo] = solver.newVar();
}

/**
 * Generate a list of all pair of directly connected station
 */
void generateAllPairStationList() {
    std::vector<int> pairStation;
    for (int stationFrom = 0; stationFrom < STATION; stationFrom++) {
        for (int stationTo = 0; stationTo < STATION; stationTo++) {
            pairStation.clear();
            if (stationFrom != stationTo && isConnected(stationFrom, stationTo)) {
                pairStation.push_back(stationFrom);
                pairStation.push_back(stationTo);
                allPairStationList.push_back(pairStation);
            }
        }
    }
}

/**
 * Generate all sets of train possible with capacity as size
 * @param capacity
 * @return
 */
std::vector<std::vector<int>> generateAllSetTrainStation(int capacity) {
    std::vector<bool> v(TRAIN);
    std::vector<std::vector<int>> result;
    for (int i = 0; i < TRAIN; ++i) {
        v[i] = (i >= (TRAIN - capacity));
    }
    do {
        std::vector<int> set;
        for (int i = 0; i < TRAIN; ++i) {
            if (v[i]) {
                set.push_back(i);
            }
        }
        result.push_back(set);
    } while (std::next_permutation(v.begin(), v.end()));
    return result;
}

/**
 * generate 3 lists of trains 1 for all the small trains, 1 for all the fast trains and 1 for all trains
 */
void nameAllTrain() {
    for (int i = 0; i < TRAIN; i++) {
        if (slowTrainsList.size() < SLOW) {
            slowTrainsList.push_back(i);
        } else {
            fastTrainsList.push_back(i);
        }
        allTrains.push_back(i);
    }
}


/**
 * Pour toute paire de gare (g, g'), pour chaque fenêtre horaire de durée TimeWindow,
 * il existe un train qui dessert g dans cette fenêtre, puis, plus tard, ce même train
 * desservira g 0 , après une durée de trajet d’au plus TravelDuration.
 */
void contrainte1() {
    literals.clear();
    for (int stationOne = 0; stationOne < STATION; stationOne++) {
        for (int stationTwo = 0; stationTwo < STATION; stationTwo++) {
            if (stationOne != stationTwo) {
                for (int iStart = 0; iStart < TIMESLOT - TIMEWINDOW; iStart = iStart + TIMEWINDOW) {
                    for (int i = iStart; i < iStart + TIMEWINDOW && i < TIMESLOT; i++) {
                        for (int i2 = i + 1; i2 <= i + TIMEDURATION && i2 < TIMESLOT; i2++) {
                            for (int train : allTrains) {
                                literals.push(Lit(J[train][i][i2][stationOne][stationTwo]));
                            }

                        }
                    }
                    solver.addClause(literals);
                    literals.clear();
                }
            }
        }
    }

}


/**
 * Pour prévenir tout risque de collision, une voie de chemin de fer entre deux gares,
 * appelé segment, peut accueillir au plus un train.
 */
void contrainte2() {
    literals.clear();
    for (int trainOne : allTrains) {
        for (int trainTwo: allTrains) {
            if (trainOne != trainTwo) {
                for (int time = 0; time < TIMESLOT; time++) {
                    for (std::vector<int> pairStation : allPairStationList) {
                        int stationFrom = pairStation[0];
                        int stationTo = pairStation[1];
                        solver.addBinary(~Lit(V[trainOne][time][stationFrom][stationTo]),
                                         ~Lit(V[trainTwo][time][stationFrom][stationTo]));
                    }
                }
            }
        }
    }
}

/**
 * Les trains de type slow sont des omnibus : ils sont tenus de s’arrêter dans toutes les
 * gares. Les trains de type fast ne sont tenus de s’arrêter que dans les gares de type big.
 */
void contrainte3() {
    literals.clear();
    for (int train : allTrains) {
        for (int time = 0; time < TIMESLOT - 1; time++) {
            for (std::vector<int> trackOne : allPairStationList) {
                for (std::vector<int> trackTwo : allPairStationList) {
                    if (trackOne[1] == trackTwo[0]) {
                        if (!(isFastTrain(train) && isSmallStation(trackOne[1]))) {
                            solver.addBinary(~Lit(V[train][time][trackOne[0]][trackOne[1]]),
                                             ~Lit(V[train][time + 1][trackTwo[0]][trackTwo[1]]));
                        }
                    }
                }
            }
        }
    }
}


/**
 * Afin de permettre aux usagers de monter à bord, tout train desservant une gare doit
 * le faire durant une durée d’au moins TimeWait.
 */
void contrainte4() {
    literals.clear();
    for (int train : allTrains) {
        for (int station = 0; station < STATION; ++station) {
            for (int timeOne = 1; timeOne < TIMESLOT; timeOne++) {
                for (int timeTwo = timeOne + 1; timeTwo < timeOne + TIMEWAIT && timeTwo < TIMESLOT; timeTwo++) {
                    literals.push(Lit(G[train][timeOne - 1][station]));
                    literals.push(~Lit(G[train][timeOne][station]));
                    literals.push(Lit(G[train][timeTwo][station]));
                    solver.addClause(literals);
                    literals.clear();
                }
            }
        }
    }

}

/**
 * Les trains doivent respecter les temps de trajet prévus par l’infrastructure entre deux
 * gares. On supposera constant, les temps de trajet entre deux gares, ils seront donc
 * donnés en entrée du problème.
 */
void contrainte5() {
    literals.clear();
    for (int train : allTrains) {
        for (std::vector<int> pairStation : allPairStationList) {
            int stationFrom = pairStation[0];
            int stationTo = pairStation[1];
            int journeyTime = map->get_duration(pairStation[0], pairStation[1]);
            for (int timeOne = 1; timeOne < TIMESLOT - journeyTime; timeOne++) {
                for (int timeTwo = timeOne + 1; timeTwo < timeOne + journeyTime && timeTwo < TIMESLOT; timeTwo++) {
                    literals.push(Lit(V[train][timeOne - 1][stationFrom][stationTo]));
                    literals.push(~Lit(V[train][timeOne][stationFrom][stationTo]));
                    literals.push(Lit(V[train][timeTwo][stationFrom][stationTo]));
                    solver.addClause(literals);
                    literals.clear();
                }
            }
        }
    }
}

/**
 * Une gare ne peut accueillir plus de trains que ce que sa capacité le permet.
 */
void contrainte6() {
    literals.clear();
    for (int station = 0; station < STATION; station++) {
        if (map->get_capacity(station) < TRAIN) {
            std::vector<std::vector<int>> allSetOfTrainForStation = generateAllSetTrainStation(
                    map->get_capacity(station) + 1);
            for (std::vector<int> setOfTrain: allSetOfTrainForStation) {
                for (int time = 0; time < TIMESLOT; ++time) {
                    for (int train : setOfTrain) {
                        literals.push(~Lit(G[train][time][station]));
                    }
                    solver.addClause(literals);
                    literals.clear();
                }
            }
        }
    }
}

/**
 * Tout train doit ˆetre quelque part
 */
void contrainteImplicite1() {
    literals.clear();
    for (int train = 0; train < TRAIN; train++) {
        for (int time = 0; time < TIMESLOT; time++) {

            /** Première partie de la clause */
            for (int station = 0; station < STATION; station++) {
                literals.push(Lit(G[train][time][station]));
            }

            /** Deuxieme partie de la clause */
            for (std::vector<int> track : allPairStationList) {
                literals.push(Lit(V[train][time][track[0]][track[1]]));
            }
            solver.addClause(literals);
            literals.clear();
        }
    }
}

/**
 * Un train ne peut pas être dans 2 gares différentes au même moment
 */
void contrainteImplicite2() {
    literals.clear();
    for (int train : allTrains) {
        for (int time = 0; time < TIMESLOT; time++) {
            for (int stationOne = 0; stationOne < STATION; stationOne++) {
                for (int stationTwo = 0; stationTwo < STATION; stationTwo++) {
                    if (stationOne != stationTwo) {
                        solver.addBinary(~Lit(G[train][time][stationOne]),
                                         ~Lit(G[train][time][stationTwo]));
                    }
                }
            }
        }
    }
}


/**
 * Un train ne peut pas être sur 2 segment différents au même moment
 */
void contrainteImplicite3() {
    literals.clear();
    for (int train : allTrains) {
        for (int time = 0; time < TIMESLOT; time++) {
            for (std::vector<int> trackOne : allPairStationList) {
                for (std::vector<int> trackTwo : allPairStationList) {
                    if (trackOne[0] != trackTwo[0] || trackOne[1] != trackTwo[1]) {
                        solver.addBinary(~Lit(V[train][time][trackOne[0]][trackOne[1]]),
                                         ~Lit(V[train][time][trackTwo[0]][trackTwo[1]]));
                    }
                }
            }
        }
    }
}

/**
 * Un train ne peut pas être sur un segment et une gare au même moment
 */
void contrainteImplicite4() {
    literals.clear();
    for (int station = 0; station < STATION; station++) {
        for (auto track : allPairStationList) {
            for (int train : allTrains) {
                for (int time = 0; time < TIMESLOT; time++) {
                    solver.addBinary(~Lit(G[train][time][station]), ~Lit(V[train][time][track[0]][track[1]]));
                }
            }
        }
    }
}

/**
 * Lorsqu'on sort d'un segment A-B, soit on arrive à gare B soit on est sur un segment B-C
 */
void contrainteImplicite5() {
    literals.clear();
    for (int train : allTrains) {
        for (int time = 1; time < TIMESLOT; time++) {
            for (std::vector<int> track : allPairStationList) {
                /** Première partie de la clause */

                int stationOne = track[0];
                int stationTwo = track[1];

                literals.push(~Lit(V[train][time - 1][stationOne][stationTwo]));
                literals.push(Lit(V[train][time][stationOne][stationTwo]));
                literals.push(Lit(G[train][time][stationTwo]));

                /** Deuxieme partie de la clause */
                for (int stationThree = 0; stationThree < STATION; stationThree++) {
                    if (isConnected(stationTwo, stationThree)) {

                        literals.push(Lit(V[train][time][stationTwo][stationThree]));
                    }
                }
                solver.addClause(literals);
                literals.clear();
            }
        }
    }
}


/**
 * Lorsqu'un train sort d'une gare A, il doit être sur un segment qui part de A
 */
void contrainteImplicite6() {
    literals.clear();
    for (int train : allTrains) {
        for (int time = 1; time < TIMESLOT; time++) {
            /** Première partie de la clause */
            for (int stationOne = 0; stationOne < STATION; stationOne++) {
                literals.push(~Lit(G[train][time - 1][stationOne]));
                literals.push(Lit(G[train][time][stationOne]));

                /**
                 * Deuxieme partie de la clause */
                for (int stationTwo = 0; stationTwo < STATION; stationTwo++) {
                    if (isConnected(stationOne, stationTwo)) {
                        literals.push(Lit(V[train][time][stationOne][stationTwo]));
                    }
                }
                solver.addClause(literals);
                literals.clear();
            }
        }
    }
}

/**
 * vérifier que  un déplacement d'une gare 1 à 2 en un temps TD correpond bien a un voyage
 */
void contrainteImplicite7() {
    literals.clear();
    for (int train : allTrains) {
        for (int stationOne = 0; stationOne < STATION; stationOne++) {
            for (int stationTwo = 0; stationTwo < STATION; stationTwo++) {
                if (stationOne != stationTwo) {
                    for (int i = 0; i < TIMESLOT - 1; i++) {
                        for (int i2 = i + 1; i2 < TIMESLOT; i2++) {
                            literals.push(~Lit(J[train][i][i2][stationOne][stationTwo]));
                            literals.push(Lit(G[train][i][stationOne]));
                            solver.addClause(literals);
                            literals.clear();
                            literals.push(~Lit(J[train][i][i2][stationOne][stationTwo]));
                            literals.push(Lit(G[train][i2][stationTwo]));
                            solver.addClause(literals);
                            literals.clear();
                        }
                    }
                }
            }
        }
    }
}

/**
 * vérifier que un voyage correpond bien a un déplacement d'une gare 1 à 2 en un temps TD
 */
void contrainteImplicite8() {
    literals.clear();
    for (int train : allTrains) {
        for (int stationOne = 0; stationOne < STATION; stationOne++) {
            for (int stationTwo = 0; stationTwo < STATION; stationTwo++) {
                if (stationOne != stationTwo) {
                    for (int i = 0; i < TIMESLOT - 1; i++) {
                        for (int i2 = i + 1; i2 < TIMESLOT; i2++) {
                            literals.push(~Lit(G[train][i][stationOne]));
                            literals.push(~Lit(G[train][i2][stationTwo]));
                            literals.push(Lit(J[train][i][i2][stationOne][stationTwo]));
                            solver.addClause(literals);
                            literals.clear();
                        }
                    }
                }
            }
        }
    }
}


/**
 * Print the result in a beautyfull way
 */
void printRes() {
    std::cout << "TIMEWAIT " << TIMEWAIT << std::endl;
    std::cout << "SLOW " << SLOW << std::endl;
    std::cout << "FAST " << FAST << std::endl;
    std::cout << "TRAIN " << TRAIN << std::endl;
    std::cout << "TIMESLOT " << TIMESLOT << std::endl;
    std::cout << "STATION " << STATION << std::endl;
    std::cout << "TRAVELDURATION " << TIMEDURATION << std::endl;
    std::cout << "TIMEWINDOW " << TIMEWINDOW << std::endl << std::endl;

    for (int t = 0; t < TRAIN; ++t) {
        std::cout << "Train number " << t;
        if (isFastTrain(t)) {
            std::cout << " (fast)";
        } else {
            std::cout << " (slow)";
        }
        std::cout << std::endl;
        for (int g = 0; g < STATION; ++g) {
            std::cout << map->get_name(g);
            std::cout << " |";
            for (int i = 0; i < TIMESLOT; ++i) {
                if (solver.model[G[t][i][g]] == l_True) {
                    std::cout << "X";
                } else {
                    std::cout << "-";
                }
            }
            std::cout << '|' << std::endl;
        }
        std::cout << std::endl;
    }
}


/**
 * Main function, get the sat solver to run
 * @return
 */
int main() {
    // ---------- Map ---------- //

    map = Graph::parse((char *) FILE);

    // ---------- Variables ---------- //

    nameAllTrain(); // nomme les trains de dispo avec un numéro entre 0 et nombre de train - 1
    generateAllPairStationList(); // génère toutes les paires (g,g')
    init_VarTrainInStationAtTime(); // Variable SAT
    init_VarTrainOnTrackAtTime(); // Variable SAT
    init_VarJourney();


    // ---------- Constraints ---------- //

    contrainte1();
    contrainte2();
    contrainte3();
    contrainte4();
    contrainte5();
    contrainte6();

    contrainteImplicite1();
    contrainteImplicite2();
    contrainteImplicite3();
    contrainteImplicite4();
    contrainteImplicite5();
    contrainteImplicite6();
    contrainteImplicite7();
    contrainteImplicite8();


    // ---------- Solver ---------- //

    printf("\n\n");
    solver.solve();
    printf("\n");



    // ---------- Printer ---------- //

    if (!solver.okay()) {
        printf("\nNO\n");
    } else {
        printRes();
    }



    // ---------- Delete ---------- //

    delete map;
    return EXIT_SUCCESS;
}
